/*
 This file is part of GNU Taler
 (C) 2022 Taler Systems S.A.

 GNU Taler is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
 Foundation; either version 3, or (at your option) any later version.

 GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
 WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with
 GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
 */

import { Logger, parseWithdrawUri } from "@gnu-taler/taler-util";
import {
  HttpResponsePaginated,
  useTranslationContext,
} from "@gnu-taler/web-util/lib/index.browser";
import { Fragment, h, VNode } from "preact";
import { Loading } from "../components/Loading.js";
import { PageStateType } from "../context/pageState.js";
import { useWithdrawalDetails } from "../hooks/access.js";
import { QrCodeSection } from "./QrCodeSection.js";
import { WithdrawalConfirmationQuestion } from "./WithdrawalConfirmationQuestion.js";

const logger = new Logger("WithdrawalQRCode");

interface Props {
  account: string;
  withdrawalId: string;
  talerWithdrawUri: string;
  onError: (e: PageStateType["error"]) => void;
  onAborted: () => void;
  onConfirmed: () => void;
  onLoadNotOk: <T>(
    error: HttpResponsePaginated<T, SandboxBackend.SandboxError>,
  ) => VNode;
}
/**
 * Offer the QR code (and a clickable taler://-link) to
 * permit the passing of exchange and reserve details to
 * the bank.  Poll the backend until such operation is done.
 */
export function WithdrawalQRCode({
  account,
  withdrawalId,
  talerWithdrawUri,
  onConfirmed,
  onAborted,
  onError,
  onLoadNotOk,
}: Props): VNode {
  const { i18n } = useTranslationContext();

  const result = useWithdrawalDetails(account, withdrawalId);
  if (!result.ok) {
    return onLoadNotOk(result);
  }
  const { data } = result;

  logger.trace("withdrawal status", data);
  if (data.aborted) {
    // signal that this withdrawal is aborted
    // will redirect to account info
    onAborted();
    return <Loading />;
  }

  const parsedUri = parseWithdrawUri(talerWithdrawUri);
  if (!parsedUri) {
    onError({
      title: i18n.str`The Withdrawal URI is not valid: "${talerWithdrawUri}"`,
    });
    return <Loading />;
  }

  if (!data.selection_done) {
    return (
      <QrCodeSection
        talerWithdrawUri={talerWithdrawUri}
        onAborted={onAborted}
      />
    );
  }

  // Wallet POSTed the withdrawal details!  Ask the
  // user to authorize the operation (here CAPTCHA).
  return (
    <WithdrawalConfirmationQuestion
      withdrawalId={parsedUri.withdrawalOperationId}
      onError={onError}
      onConfirmed={onConfirmed}
      onAborted={onAborted}
    />
  );
}
